from abc import abstractmethod
from typing import Self
from uuid import uuid4
from .callback_data import CallbackData
from .error_info import check_bad_text_and_len, check_bad_value

class Button:
    def __init__(self, text: str, callback_data: CallbackData, url: str = None):
        check_bad_text_and_len(text, self, "text")
        check_bad_value(callback_data, CallbackData, self, "callback_data")
        if url:
            check_bad_text_and_len(url, self, "url")
        
        self.text = text
        self.callback_data = callback_data
        self.url = url
    
    def clone(self) -> Self:
        return Button(self.text, self.callback_data.clone())

    def __repr__(self):
        return f"{type(self).__name__}({self.text!r}, {self.callback_data!r}, url={self.url!r})"

    def __eq__(self, other: Self):
        return (self.text == other.text and \
            self.callback_data == other.callback_data)

class ButtonRow:
    def __init__(self, *buttons: Button):
        self.buttons: list[Button] = []
        self.extend(buttons)
    
    def extend(self, buttons: list[Button]):
        for button in buttons:
            self.append(button)
        return self

    def append(self, button: Button):
        check_bad_value(button, Button, self, "button")
        self.buttons.append(button)
        return self
    
    def clone(self) -> Self:
        return ButtonRow().\
            extend([button.clone() for button in self.buttons])
    
    def __repr__(self):
        return f"{type(self).__name__}(*{self.buttons!r})"
    
    def __eq__(self, other: Self):
        return all([
            button1 == button2 
            for button1, button2
            in zip(self.buttons, other.buttons)])

class ButtonRows:
    def __init__(self, *rows: ButtonRow):
        self.rows: list[ButtonRow] = []
        self.extend(rows)
    
    def extend(self, rows: list[ButtonRow]):
        for row in rows:
            self.append(row)
    
    def append(self, row: ButtonRow):
        check_bad_value(row, ButtonRow, self, "row")
        self.rows.append(row)
    
    def clone(self) -> Self:
        return ButtonRows(*[row.clone() for row in self.rows])
    
    def __eq__(self, other: Self):
        return all([row1 == row2
            for row1, row2 in zip(self.rows,other.rows)])
        
    def __repr__(self):
        return f"{type(self).__name__}(*{self.rows!r})"
    
    @abstractmethod
    def to_reply_markup(self): ...
    
    def get_callback_data(self):
        result: list[CallbackData] = []
        for row in self.rows:
            for button in row.buttons:
                result.append(button.callback_data)
        return result


from telegram import Message as PTBMessage

from .message import HasButtonRows, Message, SentMessage
from ...message import DocumentMessage  as BaseDocumentMessage
from ...message import SentDocumentMessage  as BaseSentDocumentMessage


class DocumentMessage(BaseDocumentMessage, Message): ...

class SentDocumentMessage(BaseSentDocumentMessage, HasButtonRows, SentMessage): ...


from telegram import Message as PTBMessage

from .message import HasButtonRows, Message, SentMessage
from ...message import VideoMessage     as BaseVideoMessage
from ...message import SentVideoMessage     as BaseSentVideoMessage


class VideoMessage(BaseVideoMessage, HasButtonRows, Message): ...

class SentVideoMessage(BaseSentVideoMessage, HasButtonRows, SentMessage): ...


from abc import ABC, abstractmethod
from telegram import Bot, Message as PTBMessage

from ..button_rows import ButtonRows
from ...callback_data import CallbackDataMapping
from ...message import Message as BaseMessage
from ...message import SentMessage as BaseSentMessage

class HasButtonRows(ABC):
    def __init__(self):
        self.button_rows: ButtonRows = None
    def get_reply_markup(self, mapping: CallbackDataMapping):
        if self.button_rows:
            return self.button_rows.to_reply_markup(mapping)
        return None


class Message(BaseMessage): ...

class SentMessage(BaseSentMessage):
    def __init__(self):
        self.ptb_message: PTBMessage = None
    
    async def delete(self, bot: Bot):
        await bot.delete_message(
            chat_id=self.ptb_message.chat_id,
            message_id=self.ptb_message.message_id)


from abc import abstractmethod, ABC
from typing import Callable

from .session import InputSession

from .screen import DynamicScreen
from .input_callback import FuncCallback, InputCallback, ScreenCallback
from .callback_data import GoToScreen, RunFunc, StepBack, CallbackData
from .user_data import UserDataManager
from .user_screen import UserScreen
from .message import Message

class BotManager(ABC):
    def __init__(self):
        self.system_user_data: UserDataManager = None
        self.screen: UserScreen = None
        
    def config_delete_old_messages(self, user_id: int):
        input_callback = self.get_system_user_data(user_id).input_callback
        if input_callback is not None:
            return False
        return True
    
    def build(self):
        user_data = UserDataManager()
        screen = UserScreen(user_data)
        self.system_user_data = user_data
        self.screen = screen
        return self
    
    @abstractmethod
    def add_handlers(self): ...
    
    def get_system_user_data(self, user_id: int):
        return self.system_user_data.get(user_id)

    @abstractmethod
    def get_message_handler(self): ...

    async def _handle_message(self, user_id: int, **kwargs):
        user_data = self.get_system_user_data(user_id)
        delete_old: bool = self.config_delete_old_messages(user_id)
        if delete_old:
            await self.delete_message(**kwargs)
        
        input_callback: InputCallback = user_data.input_callback
        if input_callback is None:
            return
        
        await self.screen.clear(user_id, delete_old)
        
        for session in user_data.input_sessions:
            if not session.add_new_messages:
                continue
            message = kwargs["message"]
            session.append(message)
        
        if isinstance(input_callback, FuncCallback):
            if input_callback.one_time:
                user_data.input_callback = None
            await input_callback.function(user_id=user_id
                , **input_callback.kwargs, **kwargs)
        elif isinstance(input_callback, ScreenCallback):
            user_data.input_callback = None
            await self.screen.set_by_name(user_id, input_callback.screen_name,
                input_callback.stack, **kwargs)
            

    @abstractmethod
    def get_callback_query_handler(self): ...
    
    @abstractmethod
    async def delete_message(self, message): ...
    
    async def mapping_key_error(self, user_id: int): ...
    
    async def _handle_callback_query(self, user_id: int, query_data: str):
        user_data = self.get_system_user_data(user_id)
        mapping = user_data.callback_mapping
        data: CallbackData = mapping.get_by_uuid(query_data)
        if data is None:
            await self.mapping_key_error(user_id)
            return
        
        if isinstance(data, GoToScreen):
            if data.pre_func:
                await data.pre_func(user_id=user_id)
            
            await self.screen.set_by_name(user_id, data.screen_name)
            
            if data.post_func:
                await data.post_func(user_id=user_id)
                
            user_data.update_sessions()
        
        elif isinstance(data, StepBack):
            if data.clear_input_callback:
                user_data.input_callback = None
            
            for session in user_data.input_sessions:
                if data.pop_last_input and session.may_pop_last_input:
                    for i in range(data.times):
                        if session.messages == []:
                            break
                        session.messages.pop()
            
            if data.pre_func:
                await data.pre_func(user_id=user_id)
            
            await self.screen.step_back(user_id, data.times)
            
            if data.post_func:
                await data.post_func(user_id=user_id)
            
            user_data.update_sessions()
            
        elif isinstance(data, RunFunc):
            await data.function(user_id=user_id, **data.kwargs)
    
    def dynamic_screen(self, name: str):
        def decorator(func: Callable[[int],list[Message]]):
            self.screen.append_screen(DynamicScreen(name, func))
        return decorator


    
    



from typing import Any, Type

from .input_callback import FuncCallback

def check_bad_value(arg_value: Any, expected_type: Type, obj: Any, arg_name: str):
    if not isinstance(arg_value, expected_type):
        raise ValueError(f"У {type(obj).__name__} аргумент {arg_name}={arg_value!r} \
неправильного типа {type(arg_value).__name__} (Ожидался {expected_type.__name__})")

def check_bad_text(arg_value: str, obj: Any, arg_name: str):
    check_bad_value(arg_value, str, obj, arg_name)
        
def check_bad_text_and_len(arg_value: str, obj: Any, arg_name: str):
    check_bad_text(arg_value, obj, arg_name)
    if len(arg_value) == 0:
        raise ValueError(f"У {obj!r} аргумент {arg_name}={arg_value!r} не может быть {""!r}")

def check_pre_post_func(pre: FuncCallback, post: FuncCallback, obj: Any):
    if pre:
        check_bad_value(pre, FuncCallback, obj, "pre_func")
    
    if post:
        check_bad_value(post, FuncCallback, obj, "post_func")